# Copyright (c) Meta Platforms, Inc. and affiliates.
import vpdq  # type: ignore
import typing as t


def read_file_to_hash(input_hash_filename: str) -> t.List[vpdq.VpdqFeature]:
    """Read hash file and return vpdq hash

    Args:
        input_hash_filename (str): Input hash file path

    Returns:
        list of VpdqFeature: vpdq hash from the hash file"""

    hash = []
    with open(input_hash_filename, "r") as file:
        lines = file.readlines()
    for line in lines:
        line = line.strip()
        content = line.split(",")
        pdq_hash = vpdq.str_to_hash(content[2])
        feature = vpdq.VpdqFeature(
            int(content[1]), int(content[0]), pdq_hash, float(content[3])
        )
        hash.append(feature)

    return hash



# Copyright (c) Meta Platforms, Inc. and affiliates.

import vpdq
import subprocess
import os
import argparse
from pathlib import Path
from contextlib import contextmanager
import time

VPDQ_MATCH_DIST = 31
DIR = Path(__file__).parents[2]
VIDEOS = DIR / "tmk/sample-videos"
OUTPUT = Path(__file__).parent / "hashes"
CPP_EXEC = Path(__file__).parents[1] / "cpp/build/apps/vpdq-hash-video"


@contextmanager
def timer(context: str, print_on_enter: bool = False):
    if print_on_enter:
        print(f"{context}...")
    start = time.perf_counter()
    end = start
    yield lambda: end - start
    end = time.perf_counter()
    print(f"{context}: {end - start:.4f}s")


def get_argparse() -> argparse.ArgumentParser:
    ap = argparse.ArgumentParser()
    ap.add_argument(
        "-r",
        "--secondsPerHash",
        help="The frequence(per second) a hash is generated from the video. If it is 0, will generate every frame's hash",
        default=0,
        type=float,
    )
    ap.add_argument(
        "-s",
        "--downsampleFrameDimension",
        metavar="Downsample_Frame_Dimension",
        help="Resolution to downsample the video to before hashing frames.. If it is 0, will use the original dimension of the video to hash",
        default=0,
        type=int,
    )
    return ap


def run_benchmark(
    secondsPerHash: int,
    downsampleFrameDimension: int,
):
    with timer("Python hashing time") as pt:
        python_features = run_python(secondsPerHash, downsampleFrameDimension)
    feature_size = 0
    feature_size = sum(len(feature) for feature in python_features)
    python_time = pt()
    print(
        "  Total hash:",
        feature_size,
        "Per hash:",
        f"{1000*python_time /  feature_size:.4f}ms",
    )
    with timer("CPP hashing time") as ct:
        run_cpp(secondsPerHash, downsampleFrameDimension)
    cpp_time = ct()
    print(
        "  Total hash:",
        feature_size,
        "Per hash:",
        f"{1000*cpp_time /  feature_size:.4f}ms",
    )
    if downsampleFrameDimension != 0:
        print(
            "Calculate deviations in downsampled hashes, since hash resolution is non-native"
        )
        with timer("Original resolution Python hashing time"):
            original_features = run_python(secondsPerHash, 0)
        total_dist = 0
        mis_match = 0
        for original_hash, downsample_hash in zip(original_features, python_features):
            for original_frame, downsample_frame in zip(original_hash, downsample_hash):
                dist = original_frame.hamming_distance(downsample_frame)
                if dist >= VPDQ_MATCH_DIST:
                    mis_match += 1
                total_dist += dist
        print(
            f"  Number of mismatches: {mis_match}, {mis_match/ feature_size:.2f} percent in total."
        )
        print(f"  Average {total_dist/ feature_size:.2f} hamming distance away.")


def run_python(
    secondsPerHash: int,
    downsampleFrameDimension: int,
):
    hash_list = []
    for file in os.listdir(VIDEOS):
        if file.endswith(".mp4"):
            hash_list.append(
                vpdq.computeHash(
                    str(VIDEOS / file),
                    seconds_per_hash=secondsPerHash,
                    downsample_width=downsampleFrameDimension,
                    downsample_height=downsampleFrameDimension,
                )
            )
    return hash_list


def run_cpp(
    secondsPerHash: int,
    downsampleFrameDimension: int,
):
    Path.mkdir(OUTPUT, exist_ok=True)
    for file in os.listdir(VIDEOS):
        if file.endswith(".mp4"):
            subprocess.check_call(
                [
                    str(CPP_EXEC),
                    "-r",
                    str(secondsPerHash),
                    "-d",
                    str(OUTPUT),
                    "-s",
                    str(downsampleFrameDimension),
                    "-i",
                    str(VIDEOS) + "/" + file,
                ],
            )


def main():
    ap = get_argparse()
    ns = ap.parse_args()
    run_benchmark(**ns.__dict__)


if __name__ == "__main__":
    main()



# Copyright (c) Meta Platforms, Inc. and affiliates.

import subprocess
import os
import sys
import argparse
from pathlib import Path

VPDQ_BIN_DIR = Path(__file__).parent.parent
MATCH_HASHES_BRUTE_BIN = VPDQ_BIN_DIR / "build/cpp/apps/match-hashes-brute"


def get_argparse() -> argparse.ArgumentParser:
    ap = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter
    )

    ap.add_argument(
        "-f",
        "--queryHashFolder",
        metavar="QUERY_HASH_Folder_PATH",
        help="Query Hashes Folder's Name",
        type=dir_path,
        required=True,
    )
    ap.add_argument(
        "-i",
        "--targetHashFile",
        metavar="TARGET_HASH_FILE_PATH",
        help="Target Hash file path for comparing with the query hashes",
        required=True,
    )

    ap.add_argument(
        "-t",
        "--matchDistanceTolerance",
        metavar="Matching_distanceTolerance",
        help="The hamming distance tolerance of between two frames. If the hamming distance is bigger than the tolerance, it will be considered as unmatched",
        default="31",
        type=int,
    )
    ap.add_argument(
        "-q",
        "--qualityTolerance",
        metavar="Matching_qualityTolerance",
        help="The quality tolerance of matching two frames. If either frames is below this quality level then they will not be compared",
        default="50",
        type=int,
    )
    ap.add_argument(
        "-v",
        "--verbose",
        help="If verbose, will print detailed information.",
        action="store_true",
    )
    return ap


def dir_path(string):
    if os.path.isdir(string):
        return string
    else:
        raise argparse.ArgumentTypeError(f"readable_dir: {string} is not a valid path")


def main():
    ap = get_argparse()
    args = ap.parse_args()
    targetHashFile = args.targetHashFile
    queryHashFolder = args.queryHashFolder
    verbose = args.verbose
    distanceTolerance = str(args.matchDistanceTolerance)
    qualityTolerance = str(args.qualityTolerance)
    for file in os.listdir(queryHashFolder):
        if file.endswith(".txt"):
            print(
                "\nMatching Target:"
                + targetHashFile
                + "\n Query: "
                + f"{queryHashFolder}/{file}"
            )
            query_file = f"{queryHashFolder}/{file}"
            if verbose:
                subprocess.call(
                    [
                        MATCH_HASHES_BRUTE_EXE,
                        "-v",
                        query_file,
                        targetHashFile,
                        distanceTolerance,
                        qualityTolerance,
                    ]
                )
            else:
                subprocess.call(
                    [
                        MATCH_HASHES_BRUTE_EXE,
                        query_file,
                        targetHashFile,
                        distanceTolerance,
                        qualityTolerance,
                    ]
                )


if __name__ == "__main__":
    main()



# Copyright (c) Meta Platforms, Inc. and affiliates.

"""
Script to benchmark performance of vpdq brute_force, index with faiss and faiss only
"""

import argparse
import time

from enum import Enum
from contextlib import contextmanager, nullcontext
from threatexchange.extensions.vpdq.vpdq_brute_matcher import match_VPDQ_hash_brute
from threatexchange.extensions.vpdq.tests.utils import get_random_vpdq_features
from threatexchange.extensions.vpdq.vpdq_faiss import VPDQHashIndex
from threatexchange.extensions.vpdq.vpdq_util import (
    vpdq_to_json,
    VPDQ_QUALITY_THRESHOLD,
    VPDQ_DISTANCE_THRESHOLD,
)

import typing as t
import random
from threatexchange.extensions.vpdq.vpdq_index import VPDQIndex


class IndexType(Enum):
    BRUTE_FORCE = "brute_force"
    FLAT = "flat"
    SIGNAL_TYPE = "signal_type"

    def __str__(self) -> str:
        return self.value


@contextmanager
def timer(context: str, print_on_enter: bool = False):
    if print_on_enter:
        print(f"{context}...")
    start = time.perf_counter()
    end = start
    yield lambda: end - start
    end = time.perf_counter()
    print(f"{context}: {end - start:.4f}s")


def run_benchmark(
    test_type: IndexType,
    average_frames: int,
    jitter_noise: int,
    dataset_size: int,
    query_size: int,
):
    assert jitter_noise <= average_frames
    assert average_frames > 0
    assert dataset_size > 0
    assert query_size > 0

    data_generation_timer = nullcontext()
    if average_frames * dataset_size > 10000:
        data_generation_timer = timer("Generating data", True)
    with data_generation_timer:
        hashes = [
            get_random_vpdq_features(
                average_frames + random.randint(-jitter_noise, jitter_noise)
            )
            for _ in range(dataset_size)
        ]
    if test_type == IndexType.SIGNAL_TYPE:
        build = lambda: build_signal(hashes)
    elif test_type == IndexType.BRUTE_FORCE:
        build = lambda: hashes
    elif test_type == IndexType.FLAT:
        build = lambda: build_flat(hashes)
    else:
        raise ValueError("Invalid test type")

    with timer("build"):
        index = build()

    query_generation_timer = nullcontext()
    if query_size > 10000:
        query_generation_timer = timer("Generating queries", True)
    with query_generation_timer:
        hq = get_random_vpdq_features(query_size)
    if test_type == IndexType.SIGNAL_TYPE:
        query = lambda: signal_match(hq, index)
    elif test_type == IndexType.BRUTE_FORCE:
        query = lambda: brute_force_match(hq, index)
    elif test_type == IndexType.FLAT:
        query = lambda: index.search_with_distance_in_result(
            hq, VPDQ_DISTANCE_THRESHOLD
        )

    with timer("query") as t:
        query()
    query_time = t()
    print(f"  Per query: {1000 * query_time / query_size:.4f}ms")


def build_flat(videos):
    index = VPDQHashIndex()
    for v in videos:
        index.add_single_video(v)
    return index


def build_signal(hashes):
    index = VPDQIndex()
    for h in hashes:
        index.add(vpdq_to_json(h), object())
    return index


def signal_match(hash, index):
    index.query(vpdq_to_json(hash))


def brute_force_match(query, hashes):
    for hash in hashes:
        match_VPDQ_hash_brute(
            query, hash, VPDQ_QUALITY_THRESHOLD, VPDQ_DISTANCE_THRESHOLD
        )


def get_argparse():
    ap = argparse.ArgumentParser(description=__doc__)
    ap.add_argument(
        "--average-frames",
        "-f",
        type=int,
        default=500,
        help="How many frames in each video",
    )
    ap.add_argument(
        "--jitter-noise",
        "-j",
        type=int,
        default=50,
        help="How many frames varies between each video in uniformal distribution",
    )
    ap.add_argument(
        "--dataset-size",
        "-v",
        type=int,
        default=2000,
        help="How many videos in the dataset",
    )
    ap.add_argument(
        "--query-size",
        "-q",
        type=int,
        default=1000,
        help="number of queries",
    )
    ap.add_argument(
        "test_type",
        choices=list(IndexType),
        type=IndexType,
        help="what type of index to test",
    )
    return ap


def main():
    ap = get_argparse()
    ns = ap.parse_args()
    run_benchmark(**ns.__dict__)


if __name__ == "__main__":
    main()



# Copyright (c) Meta Platforms, Inc. and affiliates.

import os
import subprocess
import sys
import argparse
from pathlib import Path
from tempfile import TemporaryDirectory
from shutil import copyfile
import glob
import platform
from typing import Union
import csv

DIR = Path(__file__).parent
VPDQ_DIR = DIR.parent
SAMPLE_HASHES_DIR = VPDQ_DIR / "sample-hashes"
EXEC_DIR = VPDQ_DIR / "cpp/build/apps"


def get_os() -> str:
    if platform.system() == "Windows":
        return "Windows"
    elif platform.system() == "Darwin":
        return "Darwin"
    elif platform.system() == "Linux":
        return "Linux"
    else:
        print("Unknown OS. Unexpected results may occur.")
        return "Unknown"


def get_argparse() -> argparse.ArgumentParser:
    default_input_videos_dir = VPDQ_DIR.parent / "tmk/sample-videos"

    ap = argparse.ArgumentParser(description=__doc__)
    ap.add_argument(
        "-r",
        "--secondsPerHash",
        metavar="NON_NEGATIVE_FLOAT",
        help="The frequence(per second) a hash is generated from the video. If it is 0, will generate every frame's hash",
        default="0",
        type=float,
    )
    ap.add_argument(
        "-d",
        "--outputHashFolder",
        metavar="OUTPUT_HASH_Folder_PATH",
        help="Output Hash Folder's Name",
        type=validate_path,
    )
    ap.add_argument(
        "-i",
        "--inputVideoFolder",
        metavar="INPUTPUT_VIDEO_FOLDER_PATH",
        help="Input Video Folder",
        default=default_input_videos_dir,
        type=validate_path,
    )
    ap.add_argument(
        "-s",
        "--downsampleFrameDimension",
        metavar="Downsample_Frame_Dimension",
        help="Resolution to downsample the video to before hashing frames.. If it is 0, will use the original dimension of the video to hash",
        default="0",
        type=int,
    )
    ap.add_argument(
        "-t",
        "--matchDistanceTolerance",
        metavar="Matching_distanceTolerance",
        help="The hamming distance tolerance of between two frames. If the hamming distance is bigger than the tolerance, it will be considered as unmatched",
        default="31",
        type=int,
    )
    ap.add_argument(
        "-q",
        "--qualityTolerance",
        metavar="Matching_qualityTolerance",
        help="The quality tolerance of matching two frames. If either frames is below this quality level then they will not be compared",
        default="50",
        type=int,
    )
    ap.add_argument(
        "-v",
        "--verbose",
        help="If verbose, will print detailed information.",
        action="store_true",
    )
    return ap


def validate_path(path: Union[Path, str], err_msg: Union[str, None] = None) -> Path:
    path = Path(path)
    if not path.exists():
        if err_msg is None:
            err_msg = f"Path {path} does not exist."
        raise argparse.ArgumentTypeError(err_msg)
    return path


def main():
    OS = get_os()

    hashVideoExecutable = EXEC_DIR / "vpdq-hash-video"
    matchHashesExecutable = EXEC_DIR / "match-hashes-byline"

    if OS == "Windows":
        hashVideoExecutable = Path(f"{hashVideoExecutable}.exe")
        matchHashesExecutable = Path(f"{matchHashesExecutable}.exe")

    try:
        validate_path(hashVideoExecutable)
        validate_path(matchHashesExecutable)
    except argparse.ArgumentTypeError as e:
        print(e)
        print(
            "Error: Hashing executable/s not found. Build vpdq before running regtest."
        )
        sys.exit(1)

    ap = get_argparse()
    args = ap.parse_args()
    inputVideoFolder = args.inputVideoFolder
    outputHashFolder = args.outputHashFolder
    secondsPerHash = str(args.secondsPerHash)
    downsampleFrameDimension = str(args.downsampleFrameDimension)
    distanceTolerance = str(args.matchDistanceTolerance)
    qualityTolerance = str(args.qualityTolerance)
    verbose = args.verbose

    # Run the hashing and matching tests for single and multithreaded
    for thread_count in range(0, 2):
        if thread_count == 0:
            num_cpu_cores = os.cpu_count()
            print(f"Number of hashing threads: auto. Probably {num_cpu_cores} threads.")
        else:
            print(f"Number of hashing threads: {thread_count}")
        with TemporaryDirectory() as tempOutputHashFolder:
            tempOutputHashFolder = Path(tempOutputHashFolder)

            # Create output directory if it does not exist and it is specified
            if outputHashFolder is not None:
                if not outputHashFolder.exists():
                    print(f"Creating output directory at {outputHashFolder}")
                    outputHashFolder.mkdir(parents=True)
                print(f"Writing output hash files to directory: {outputHashFolder}")
            else:
                print(f"Writing output to temp directory: {tempOutputHashFolder}")
            # TODO: Add more general options for other video extensions.
            for fileStr in glob.iglob(f"{inputVideoFolder}/**/*.mp4", recursive=True):
                file = Path(fileStr)

                # Create output hash file in a tempdir
                outputHashFile = tempOutputHashFolder / f"{file.stem}.txt"
                outputHashFile.touch(exist_ok=False)

                print(f"\nHashing file {file.name}")
                command = [
                    hashVideoExecutable,
                    "-r",
                    secondsPerHash,
                    "-d",
                    tempOutputHashFolder,
                    "-s",
                    downsampleFrameDimension,
                    "-i",
                    file,
                    "-t",
                    str(thread_count),
                ]

                if verbose:
                    # Print all PDQHashes e.g.
                    # PDQHash: ebcc8b06b0666ea34cf9b85972a983a4f94668af05fc9d52aa9662f975499514
                    # selectframe 563
                    command.insert(1, "-v")

                try:
                    hash_proc = subprocess.run(
                        command,
                        check=True,
                        capture_output=True,
                        shell=(OS == "Windows"),
                    )
                    print(str(hash_proc.stdout, "utf-8"))
                except subprocess.CalledProcessError as e:
                    print(" ".join([str(i) for i in e.cmd]))
                    print(str(e.stderr, "utf-8"))
                    sys.exit(1)

                # Copy hash files to output directory if it is specified
                # This will overwrite existing files with the same
                # name as outputHashFile in the directory
                if outputHashFolder is not None:
                    copyfile(
                        outputHashFile,
                        Path(
                            outputHashFolder
                            / f"{outputHashFile.stem}-{thread_count}thread.txt"
                        ),
                    )

            for outputFileStr in glob.iglob(
                f"{tempOutputHashFolder}/**/*.txt", recursive=True
            ):
                outputFile = Path(outputFileStr)
                sampleFile = Path(SAMPLE_HASHES_DIR / outputFile.name)
                print(
                    f"\nMatching Video {sampleFile.name} from hash file {outputFile.name}"
                )
                command = [
                    matchHashesExecutable,
                    sampleFile,
                    outputFile,
                    distanceTolerance,
                    qualityTolerance,
                ]

                if verbose:
                    # Print all PDQHashes and if they match e.g.
                    # Line 201 Hash1: da4b380330b725b4a5f08ff03d0f6949da4fd2d3e7c8e4930fa7b80662a17c4e
                    # Hash2: da4b380330b725b4a5f08ff03d0f6949da4fd2d3e7c8e4930fa7b80662a17c4e match
                    command.insert(1, "-v")

                try:
                    match_proc = subprocess.run(
                        command,
                        check=True,
                        capture_output=True,
                        shell=(OS == "Windows"),
                    )
                    print(str(match_proc.stdout, "utf-8"))
                except subprocess.CalledProcessError as e:
                    print(e.cmd)
                    print(str(e.stderr, "utf-8"))
                    sys.exit(1)

            # Test that all the features are in frame order
            for outputFileStr in glob.iglob(
                f"{tempOutputHashFolder}/**/*.txt", recursive=True
            ):
                outputFile = Path(outputFileStr)
                with open(outputFile, "r") as f:
                    features = csv.reader(f)
                    oldFrameNumber = -1
                    for feature in features:
                        frameNumber = int(feature[0])
                        assert frameNumber >= 0
                        assert frameNumber > oldFrameNumber
                        assert (oldFrameNumber + 1) == frameNumber
                        oldFrameNumber = frameNumber
            print("All features are in frame order.")
            print("\n--------------------------------------\n")


if __name__ == "__main__":
    main()



# distutils: language = c++
# cython: language_level=3
# Copyright (c) Meta Platforms, Inc. and affiliates.
import typing as t

from pathlib import Path
from dataclasses import dataclass
from libcpp cimport bool
from libcpp.vector cimport vector
from libcpp.string cimport string


cdef extern from "pdq/cpp/common/pdqhashtypes.h" namespace "facebook::pdq::hashing":
    cdef struct Hash256:
        unsigned short w[16];
    int hammingDistance(
        Hash256 hash1,
        Hash256 hash2
    )
    string hashToString(
        Hash256 hash
    )

cdef extern from "pdq/cpp/common/pdqhashtypes.h" namespace "facebook::pdq::hashing::Hash256":
    Hash256 fromStringOrDie(
        char* hex_str
    )

cdef extern from "vpdq/cpp/hashing/vpdqHashType.h" namespace "facebook::vpdq::hashing":
    cdef struct vpdqFeature:
        Hash256 pdqHash;
        int frameNumber;
        int quality;
        double timeStamp;


cdef extern from "vpdq/cpp/hashing/filehasher.h" namespace "facebook::vpdq::hashing":
    bool hashVideoFile(
        string input_video_filename,
        vector[vpdqFeature]& pdqHashes,
        bool verbose,
        double seconds_per_hash,
        int width,
        int height,
        unsigned int thread_count,
    )


@dataclass
class VpdqFeature:
    quality: int
    frame_number: int
    hash: Hash256
    hex: str
    timestamp: float

    def __init__(
        self, quality: int, frame_number: int, hash: Hash256, timestamp: float
    ):
        self.quality = quality
        self.frame_number = frame_number
        self.hash = hash
        self.hex = hash_to_hex(hash)
        self.timestamp = timestamp

    def hamming_distance(self, that: VpdqFeature):
        return hammingDistance(self.hash, that.hash)


def hash_to_hex(hash_value: Hash256) -> str:
    """Convert from pdq hash to hex str

    Args:
        hash_value

    Returns:
        str: hex str of hash
    """
    return str(hashToString(hash_value), "utf-8")


def str_to_hash(str_hash: str):
    return fromStringOrDie(str(str_hash).encode("utf-8"))


def hamming_distance(hash1: Hash256, hash2: Hash256) -> int:
    """
    Return the hamming distance between two pdq hashes

    Args:
        hash1
        hash2

    Returns:
        int: The hamming distance
    """
    return hammingDistance(hash1, hash2)


def computeHash(
    input_video_filename: t.Union[str, Path],
    ffmpeg_path: t.Union[str, None] = None,
    seconds_per_hash: float = 1.0,
    verbose: bool = False,
    downsample_width: int = 0,
    downsample_height: int = 0,
    thread_count: int = 0,
) -> t.List[VpdqFeature]:
    """Compute vpdq hash

    Args:
        input_video_filename: Input video file path
        ffmpeg_path: ffmpeg path (this is not used anymore)
        verbose: If verbose, will print detailed information
        seconds_per_hash: The frequence(per second) a hash is generated from the video. If it is 0, will generate every frame's hash
        downsample_width: Width to downsample the video to before hashing frames. If it is 0, will use the original width of the video to hash
        downsample_height: Height to downsample the video to before hashing frames. If it is 0, will use the original height of the video to hash
        thread_count: Number of threads for hashing. If it is 0, will use choose automatically
    Returns:
        list of vpdq_feature: VPDQ hash from the video
    """
    str_path = str(input_video_filename)
    if not Path(str_path).is_file():
        raise ValueError("Input_video_filename doesn't exist")
    if seconds_per_hash < 0:
        raise ValueError("Seconds_per_hash must be non-negative")
    if downsample_width < 0:
        raise ValueError("Downsample_width must be non-negative")
    if downsample_height < 0:
        raise ValueError("Downsample_height must be non-negative")
    if thread_count < 0:
        raise ValueError("Thread_count must be non-negative")
    cdef vector[vpdqFeature] vpdq_hash;
    

    rt = hashVideoFile(
        str_path.encode("utf-8"),
        vpdq_hash,
        verbose,
        seconds_per_hash,
        downsample_width,
        downsample_height,
        thread_count,
    )

    if not rt:
        raise Exception("Fail to create VPDQ hash")

    hashes = [
        VpdqFeature(hash.quality, hash.frameNumber, hash.pdqHash, hash.timeStamp)
        for hash in vpdq_hash
    ]
    return hashes


def _cli():
    import argparse

    ap = argparse.ArgumentParser(
        description="a simple wrapper for the vPDQ hashing algorithm"
    )
    ap.add_argument("file", help="the file to hash")
    ns = ap.parse_args()
    hashes = computeHash(ns.file)
    for hash in hashes:
        print(f"{hash.hex},{hash.quality},{hash.timestamp:.3f}")


# Copyright (c) Meta Platforms, Inc. and affiliates.
import vpdq  # type: ignore
import pytest
import test_util
from pathlib import Path
import glob
import re
from typing import Union, Sequence, NamedTuple
from collections import namedtuple

DIR = Path(__file__).parent
VPDQ_DIR = DIR.parent.parent
SAMPLE_HASH_FOLDER = VPDQ_DIR / "sample-hashes"
SAMPLE_VIDEOS = VPDQ_DIR.parent / Path("tmk/sample-videos")
DISTANCE_TOLERANCE = 31
QUALITY_TOLERANCE = 50

PROJECT_DIR = Path(__file__).parents[3]
HASH_FOLDER = PROJECT_DIR / SAMPLE_HASH_FOLDER
VIDEO_FOLDER = PROJECT_DIR / SAMPLE_VIDEOS
test_hashes = {}
sample_hashes = {}


def get_test_file_paths() -> Sequence[Path]:
    test_files = []
    for fileStr in glob.iglob(f"{VIDEO_FOLDER}/**/*.mp4", recursive=True):
        file = Path(fileStr)
        if not (VIDEO_FOLDER / f"{file.name}").is_file():
            print(f"Video file {file.name} doesn't exist. Skipping.")
            continue
        test_files.append(file)
    assert len(test_files) > 0
    return test_files


def test_vpdq_utils():
    test_files = get_test_file_paths()
    sample = Path(f"{HASH_FOLDER}/{test_files[0].stem}.txt")
    assert sample.is_file()
    ret = test_util.read_file_to_hash(sample)
    with open(sample, "r") as file:
        lines = file.readlines()
    for line, feature in zip(lines, ret):
        line = line.strip()
        content = line.split(",")
        hex_hash = content[2]
        assert vpdq.str_to_hash(hex_hash) == feature.hash
        assert vpdq.hash_to_hex(feature.hash) == hex_hash


def test_error_checking():
    test_files = get_test_file_paths()
    video_file = Path(f"{VIDEO_FOLDER}/{test_files[0].name}")
    with pytest.raises(ValueError, match="Seconds_per_hash must be non-negative"):
        vpdq.computeHash(input_video_filename=video_file, seconds_per_hash=-1)

    with pytest.raises(ValueError, match="Downsample_width must be non-negative"):
        vpdq.computeHash(input_video_filename=video_file, downsample_width=-1)

    with pytest.raises(ValueError, match="Downsample_height must be non-negative"):
        vpdq.computeHash(input_video_filename=video_file, downsample_height=-1)

    with pytest.raises(ValueError, match="Input_video_filename doesn't exist"):
        vpdq.computeHash(input_video_filename="nonexisting")


def get_downsampled_hash_files(
    input_hash_file_path: Union[Path, str]
) -> Sequence[NamedTuple]:
    """
    Get all the downsampled hash files that match the input hash file name
    The input hash file name should be in the format of <base_name>.txt
    The downsampled hash file name should be in the format of <base_name>-<width>x<height>.txt
    """

    # Define the regular expression pattern to match the desired filenames
    pattern = r"(.*)-(\d+)x(\d+)\.txt"

    # Get the base file name without resolution and extension
    base_file_name_match = re.match(r"(.*)(?=\.txt)", Path(input_hash_file_path).name)
    if base_file_name_match is not None:
        base_file_name = base_file_name_match.group(1)
    else:
        return []

    pathanddim = namedtuple("pathanddim", ["path", "width", "height"])

    all_matches = []
    for fileStr in HASH_FOLDER.iterdir():
        file = Path(fileStr)
        if file.is_file():
            matches = re.findall(pattern, file.name)
            for match in matches:
                name = match[0]
                width = match[1]
                height = match[2]
                # Only print matches that have the same base name as the input file
                if name == base_file_name:
                    new_match = pathanddim(Path(file), width, height)
                    all_matches.append(new_match)
    return all_matches


def test_compare_hashes():
    """This regression test is creating hashes from sample videos and compare them with the provided hashes line by line.
    Two VPDQ features are considered the same if each line of the hashes are within DISTANCE_TOLERANCE.
    For hashes that have a quality lower than QUALITY_TOLERANCE, the test will skip them for comparing.
    """
    test_files = get_test_file_paths()

    for file in test_files:
        # Load the hash file truth
        hash_file = Path(f"{HASH_FOLDER}/{file.stem}.txt")
        assert hash_file.is_file()
        ret = test_util.read_file_to_hash(hash_file)
        assert ret is not None
        sample_hashes[file] = ret

        # Calculate the hash of file
        assert file.is_file()
        ret = vpdq.computeHash(input_video_filename=file, seconds_per_hash=0)
        assert ret is not None
        test_hashes[file] = ret

        print("Comparing hash for video:", file)
        hash1 = test_hashes[file]
        hash2 = sample_hashes[file]
        assert len(hash1) == len(hash2)
        for h1, h2 in zip(hash1, hash2):
            if h1.quality >= QUALITY_TOLERANCE and h2.quality >= QUALITY_TOLERANCE:
                assert h1.hamming_distance(h2) < DISTANCE_TOLERANCE
                assert h1.frame_number == h2.frame_number

        # Compare the downsampled hashes
        for downsampled_hash_file_pathanddim in get_downsampled_hash_files(hash_file):
            downsampled_hash_file = downsampled_hash_file_pathanddim.path
            downsampled_width = int(downsampled_hash_file_pathanddim.width)
            downsampled_height = int(downsampled_hash_file_pathanddim.height)

            # Load the downsampled hash file truth
            assert downsampled_hash_file.is_file()
            ret = test_util.read_file_to_hash(downsampled_hash_file)
            assert ret is not None
            sample_hashes[downsampled_hash_file] = ret

            # Calculated downsampled hash of test file
            assert file.is_file()
            ret = vpdq.computeHash(
                input_video_filename=file,
                seconds_per_hash=0,
                downsample_width=downsampled_width,
                downsample_height=downsampled_height,
            )
            assert ret is not None
            test_hashes[downsampled_hash_file] = ret

            print("Comparing hash for downsampled video:", downsampled_hash_file)
            hash1 = test_hashes[downsampled_hash_file]
            hash2 = sample_hashes[downsampled_hash_file]
            assert len(hash1) == len(hash2)
            for h1, h2 in zip(hash1, hash2):
                if h1.quality >= QUALITY_TOLERANCE and h2.quality >= QUALITY_TOLERANCE:
                    assert h1.hamming_distance(h2) < DISTANCE_TOLERANCE
                    assert h1.frame_number == h2.frame_number



# Copyright (c) Meta Platforms, Inc. and affiliates.

import vpdq  # type: ignore
from pathlib import Path
from typing import Union, Sequence
import glob
import argparse

DIR = Path(__file__).parent
VPDQ_DIR = DIR.parent.parent
PROJECT_DIR = Path(__file__).parents[3]


def get_argparse() -> argparse.ArgumentParser:
    default_video_folder = PROJECT_DIR / Path("tmk/sample-videos")
    default_hash_folder = VPDQ_DIR / "sample-hashes"

    ap = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter
    )

    ap.add_argument(
        "-i",
        "--inputVideoFolder",
        help="Video folder to hash recursively",
        type=dir_path,
        default=default_video_folder,
    )

    ap.add_argument(
        "-d",
        "--outputHashFolder",
        help="Query Hashes Folder's Name",
        type=dir_path,
        default=default_hash_folder,
    )

    ap.add_argument(
        "--overwrite",
        help="Overwrite existing hashes in the folder.",
        type=bool,
        default=False,
        action=argparse.BooleanOptionalAction,
    )

    ap.add_argument(
        "-v",
        "--verbose",
        help="Print verbose output.",
        type=bool,
        default=False,
        action=argparse.BooleanOptionalAction,
    )

    return ap


def dir_path(path: Union[Path, str]) -> Path:
    path = Path(path)
    if not Path(path).is_dir():
        raise argparse.ArgumentTypeError(f"{path} is not a valid path")
    return path


def generate_hashes(
    output_folder: Path,
    input_filepaths: Sequence[Path],
    overwrite: bool = False,
    verbose: bool = False,
    downsample_width: int = 0,
    downsample_height: int = 0,
) -> int:
    """
    Generate vpdq hashes for the test videos
    and write them to files in output_folder

    If overwrite is set, it will overwrite existing hashes in the folder.

    Returns the number of files hashed and written.
    """
    hash_count = 0
    for fileStr in input_filepaths:
        file = Path(fileStr)
        if downsample_width > 0 and downsample_height > 0:
            output_hash_filename = (
                output_folder
                / f"{file.stem}-{downsample_width}x{downsample_height}.txt"
            )
        else:
            output_hash_filename = output_folder / f"{file.stem}.txt"

        if not overwrite and output_hash_filename.exists():
            if verbose:
                print(
                    f"Hash file {output_hash_filename.name} already exists. Skipping."
                )
            continue

        vpdq_hash = vpdq.computeHash(
            input_video_filename=file,
            seconds_per_hash=0,
            downsample_width=downsample_width,
            downsample_height=downsample_height,
        )

        # Overwrite existing hashes if overwrite is set
        filemode = "w+t" if overwrite else "x+t"
        with open(output_hash_filename, filemode) as output_file:
            for feature in vpdq_hash:
                output_file.write(
                    f"{feature.frame_number},{feature.quality},{vpdq.hash_to_hex(feature.hash)},{feature.timestamp:.3f}\n"
                )
        hash_count += 1
    return hash_count


def get_test_file_paths(video_folder: Path) -> Sequence[Path]:
    test_files = []
    for fileStr in glob.iglob(f"{video_folder}/**/*.mp4", recursive=True):
        file = Path(fileStr)
        if not (video_folder / f"{file.name}").is_file():
            print(f"Video file {file.name} doesn't exist. Skipping.")
            continue
        test_files.append(file)
    assert len(test_files) > 0
    return test_files


def main():
    """
    Generate hashes for the sample videos.
    This can be used to update the hashes for the sample videos if vpdq is changed.
    """
    ap = get_argparse()
    args = ap.parse_args()
    inputVideoFolder = args.inputVideoFolder
    outputHashFolder = args.outputHashFolder
    overwrite = args.overwrite
    verbose = args.verbose

    test_files = get_test_file_paths(inputVideoFolder)

    print(f"Generating hashes for videos in {inputVideoFolder}")
    print(f"Outputting hashes to {outputHashFolder}")
    print("Overwriting existing hashes.") if overwrite else print("")

    hash_count = generate_hashes(
        outputHashFolder, test_files, overwrite=overwrite, verbose=verbose
    )
    print(f"Generated {hash_count} new hashes.")

    # Generate 128x128 downsampled hashes for the sample videos
    downsampled_hash_count = generate_hashes(
        outputHashFolder,
        test_files,
        overwrite=overwrite,
        verbose=verbose,
        downsample_width=128,
        downsample_height=128,
    )

    print(f"Generated {downsampled_hash_count} new downsampled hashes.")


if __name__ == "__main__":
    main()
